use crate::entity::{Point as PointEntity, PointRef};
use crate::math::{Line2, Vec2, Point2, Region1, Region2, Rot2, Scalar};

#[derive(Clone, Copy, Debug, PartialEq)]
pub enum ResolveResult {
    Underconstrained,
    Constrained,
    Overconstrained,
}

impl ResolveResult {
    pub fn from_r2(r: &Region2) -> ResolveResult {
        use Region2::*;
        match r {
            Empty => ResolveResult::Overconstrained,
            Singleton(_) => ResolveResult::Constrained,
            _ => ResolveResult::Constrained,
        }
    }
}

pub trait Relation {
    fn resolve(&self) -> ResolveResult;
}

pub struct Coincident {
    pub p1: PointRef,
    pub p2: PointRef,
}

impl Relation for Coincident {
    fn resolve(&self) -> ResolveResult {
        let (mut p1, mut p2) = (self.p1.borrow_mut(), self.p2.borrow_mut());
        let r = { p1.pos.constraints().intersect(p2.pos.constraints()) };
        p1.pos.reconstrain(r.clone());
        p2.pos.reconstrain(r.clone());
        ResolveResult::from_r2(&r)
    }
}

pub struct PointAngle {
    pub p1: PointRef,
    pub p2: PointRef,
    pub angle: Rot2,
}

impl PointAngle {
    pub fn new(p1: PointRef, p2: PointRef, angle: Rot2) -> Self {
        Self { p1, p2, angle }
    }

    pub fn new_horizontal(p1: PointRef, p2: PointRef) -> Self {
        Self::new(p1, p2, Rot2::from_cos_sin_unchecked(1., 0.))
    }

    pub fn new_vertical(p1: PointRef, p2: PointRef) -> Self {
        Self::new(p1, p2, Rot2::from_cos_sin_unchecked(0., 1.))
    }
}

impl Relation for PointAngle {
    fn resolve(&self) -> ResolveResult {
        use Region2::*;
        let (mut p1, mut p2) = (self.p1.borrow_mut(), self.p2.borrow_mut());
        let constrain_line = |p1: &Point2, p2: &mut PointEntity| {
            let line = Region2::Line(Line2::new(*p1, self.angle, Region1::Full));
            let new_constraint = p2.pos.constraints().intersect(&line);
            p2.pos.reconstrain(new_constraint);
            ResolveResult::from_r2(p2.pos.constraints())
        };
        match (&mut p1.pos.constraints(), &mut p2.pos.constraints()) {
            (Empty, _) | (_, Empty) => ResolveResult::Overconstrained,
            (Singleton(p1), Singleton(p2)) => {
                // if the angle p1 and p2 form is parallel to self.angle, the result
                // will have a y component of 0
                let r = self.angle.to_rotation_matrix().inverse() * (p2 - p1);
                if relative_eq!(r.y, 0.) {
                    ResolveResult::Constrained
                } else {
                    ResolveResult::Overconstrained
                }
            }
            (Singleton(p), _) => constrain_line(p, &mut *p2),
            (_, Singleton(p)) => constrain_line(p, &mut *p1),
            _ => ResolveResult::Underconstrained,
        }
    }
}

pub enum Axis {
    Vertical,
    Horizontal,
}

pub struct AlignedDistance {
    pub p1: PointRef,
    pub p2: PointRef,
    pub axis: Axis,
    pub distance: Scalar,
}

impl AlignedDistance {
    pub fn new(p1: PointRef, p2: PointRef, axis: Axis, distance: Scalar) -> Self {
        Self {
            p1,
            p2,
            axis,
            distance,
        }
    }

    pub fn new_vertical(p1: PointRef, p2: PointRef, distance: Scalar) -> Self {
        Self::new(p1, p2, Axis::Vertical, distance)
    }

    pub fn new_horizontal(p1: PointRef, p2: PointRef, distance: Scalar) -> Self {
        Self::new(p1, p2, Axis::Horizontal, distance)
    }
}

impl Relation for AlignedDistance {
    fn resolve(&self) -> ResolveResult {
        use Region2::*;
        let (mut p1, mut p2) = (self.p1.borrow_mut(), self.p2.borrow_mut());
        let constrain_line = |p1: Point2, p2: &mut PointEntity| {
            let angle = match self.axis {
                Axis::Horizontal => Rot2::from_cos_sin_unchecked(0., 1.),
                Axis::Vertical => Rot2::from_cos_sin_unchecked(1., 0.),
            };
            let line = Region2::Line(Line2::new(p1, angle, Region1::Full));
            let new_constraint = p2.pos.constraints().intersect(&line);
            p2.pos.reconstrain(new_constraint);
            ResolveResult::from_r2(p2.pos.constraints())
        };
        let offset = match self.axis {
            Axis::Horizontal => Vec2::new(self.distance, 0.),
            Axis::Vertical => Vec2::new(0., self.distance),
        };
        match (&mut p1.pos.constraints(), &mut p2.pos.constraints()) {
            (Empty, _) | (_, Empty) => ResolveResult::Overconstrained,
            (Singleton(p1), Singleton(p2)) => {
                let r = p2 - p1;
                let d = match self.axis {
                    Axis::Horizontal => r.x,
                    Axis::Vertical => r.y,
                };
                if relative_eq!(d, self.distance) {
                    ResolveResult::Constrained
                } else {
                    ResolveResult::Overconstrained
                }
            }
            (Singleton(pos), _) => constrain_line(pos + offset, &mut *p2),
            (_, Singleton(pos)) => constrain_line(pos - offset, &mut *p1),
            _ => ResolveResult::Underconstrained,
        }
    }
}
